﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Text.RegularExpressions;
using System.Collections;
using System.Xml.Linq;

namespace MidiToSound
{
    class MidiToSound
    {

        /******************************************************************
        * MID2SND version 0.04 (current) 
        * 
        * Change history:
        * ---------------
        *   v 0.02 - ??/??/???? - Original application programmed by Ravi I.
        *   v 0.03 - 06/17/2012 - Ported to C# by Jeremiah Nellis
        *   v 0.04 - 06/23/2012 - Added support for specifing devices, JN 
        *******************************************************************/

        static void Main(string[] args)
            {
               ushort tempo;
               BinaryReader infile = null;
               BinaryWriter outfile = null;
               int instrument = -1;
               int fixcontrols = 0;
               int devices = 0;
               //int init = 1;   JN - Never used
               string xmlConfigFile = null;

               // check for command-line arguments
               if ( args.Length < 2 )
               {
                  DisplayHelp();
                  return;
               }

               // interpret flags

               for ( int lcv=2; lcv<args.Length; lcv++ )
               {
                  if ( args[lcv][0] == '-' )
                     switch ( args[lcv][1] )
                     {
                        case 'I':
                             instrument = Convert.ToInt16(args[lcv+1]);
                           break;

                        case 'C':
                           fixcontrols = 1;
                           break;

                        // JN - Never used
                        //case 'M':
                        //   init = 0;
                        //   break;
                        //

                        case 'D':
                           string[] deviceList = Regex.Split(Convert.ToString(args[lcv + 1]), ",");
                           for (int i = 0; i < deviceList.Length; i++)
                           {
                               if (string.Equals("MT32", deviceList[i], StringComparison.CurrentCultureIgnoreCase) || deviceList[i] == "1")
                               {
                                   devices = devices | 0x01; 
                               }
                               if (string.Equals("Yamaha", deviceList[i], StringComparison.CurrentCultureIgnoreCase) || deviceList[i] == "2")
                               {
                                   devices = devices | 0x02;
                               }
                               if (string.Equals("AdLib", deviceList[i], StringComparison.CurrentCultureIgnoreCase) || deviceList[i] == "4")
                               {
                                   devices = devices | 0x04;
                               }
                               if (string.Equals("Casio", deviceList[i], StringComparison.CurrentCultureIgnoreCase) || deviceList[i] == "8")
                               {
                                   devices = devices | 0x08;
                               }
                               if (string.Equals("Tandy", deviceList[i], StringComparison.CurrentCultureIgnoreCase) || deviceList[i] == "10")
                               {
                                   devices = devices | 0x10;
                               }
                               if (string.Equals("PC", deviceList[i], StringComparison.CurrentCultureIgnoreCase) || deviceList[i] == "20")
                               {
                                   devices = devices | 0x20;
                               }
                               if (string.Equals("Unknown", deviceList[i], StringComparison.CurrentCultureIgnoreCase) || deviceList[i] == "40")
                               {
                                   devices = devices | 0x40;
                               }
                               if (string.Equals("GM", deviceList[i], StringComparison.CurrentCultureIgnoreCase) || deviceList[i] == "80")
                               {
                                   devices = devices | 0x80;
                               }
                           }
                           break;

                         case 'X':
                           xmlConfigFile = args[lcv + 1];
                           break;

                        default:
                           Console.Write("Unknown flag {0}\n", args[lcv]);
                           break;
                     }
               }

               // open the input and output files

               try
               {
                   infile = new BinaryReader(File.Open(args[0], FileMode.Open));
               }
               catch
               {
                   ErrorExit("Error: Unable to open midi file for input\n");
               }

               try
               {
                   outfile = new BinaryWriter(File.Open(args[1], FileMode.Create));
               }
               catch
               {
                   ErrorExit("Error: Unable to open sound file for output\n");
               }


               // read the MIDI header and track header
               tempo = ReadMThd (ref infile);
               ReadMTrk(ref infile);

               // Initialize the sound resource
               WriteIdentifier(ref outfile);

               if (xmlConfigFile != null)
               {
                   WriteHeaderFromConfigFile(ref outfile, ref xmlConfigFile);
               }
               else
               {
                   // initialize channel info
                   WriteHeader(ref outfile, devices);
               }

               // initialize controls
               /* FIXME: JN - Removed.  This generates additional header data that we can't use
               if ( Convert.ToBoolean(init) )
                  InitControls(ref outfile);
               */

               // copy events
               CopyEvents(ref infile, ref outfile, tempo, instrument, fixcontrols);

               // close the open files
               infile.Close();
               outfile.Close();
            }


            //-------------------------------------------------------------------------
            // DisplayHelp
            // Called when the program is run without argumenmts
            // Displays help information

            static void DisplayHelp ()
            {
               Console.Write("\nMIDI to SCI Sound Converter\n");
               Console.Write("version 0.03\n");
               Console.Write("By Ravi I.\n\n");
               Console.Write("Ported to C# by Jeremiah Nellis - June 17, 2012\n");
               Console.Write("    Modified to include device configuration, JN - June 23, 2012\n");
               Console.Write("Usage:\n");
               Console.Write("   midi2snd <midifile> <soundfile> [flags]\n\n");
               Console.Write("<soundfile> is the filename of an extracted sound resource\n");
               Console.Write("<midifile> is the output file\n\n");
               Console.Write("Flags must come after the filenames if included. A flag\n");
               Console.Write("must be separated from its paramater with a space.\n");
               Console.Write("   -I n     - set program to n\n");
               Console.Write("   -C       - remove references to unrecognized controls\n");
               //Console.Write("   -M       - leave channels in monophonic mode\n"); JN - Not implemented
               Console.Write("   -D 1,2,4,8,10,20,40,80\n");
               Console.Write("      - Comma separated list of devices to enable (on all 16 channels)\n");
               Console.Write("        Can also use device names instead of numbers:\n");
               Console.Write("            MT32,Yamaha,AdLib,Casio,Tandy,PC,Unknown,GM\n");
               Console.Write("        For control over individual device/channel combinations, use a \n");
               Console.Write("        configuration input file instead.\n");
               Console.Write("   -X <filename>\n");
               Console.Write("      - XML configuration input file containing device/channel information.\n");
               Console.Write("        Use SoundToMidi application to generate.\n");
            }

            //-------------------------------------------------------------------------
            //   ErrorExit
            //   Called when the program encounters an error and must quit
            //   Displays an error message then exits

            static void ErrorExit (string message)
            {
               Console.Write(message);
               Environment.Exit(1);
            }


            //-------------------------------------------------------------------------
            //   WriteIdentifier
            //   Called with the output file handle to check for a sound resource
            //   Writes the first two bytes of the file as a short int

            static void WriteIdentifier (ref BinaryWriter fileh)
            {
               fileh.BaseStream.Position = 0;
               byte[] id = { 0x84, 0x00 };
               fileh.Write(id);
            }


            //-------------------------------------------------------------------------
            //   ReadChar
            //   Called with the input file handle to read a single byte
            //   Returns the value or -1 on error

            static short ReadChar (ref BinaryReader fileh)
            {
               byte? charbuf;

               charbuf = fileh.ReadBytes(1)[0];
               if(charbuf == null) {
                  return -1;
               }

               return (short)charbuf;
            }


            //-------------------------------------------------------------------------
            //   ReadVL
            //   Called with the output file handle and a data byte
            //   Reads the variable-legth value

            static long ReadVL (ref BinaryReader fileh)
            {
               byte foo;
               ulong bar = 0;
               short? baz;

               while ( (baz = fileh.ReadBytes(1)[0]) != null )
               {
                  foo = (byte)baz;
                  bar = (ulong)((ushort)(bar << 7) | (ushort)(foo & 0x7F));
                  if ( !(Convert.ToBoolean(foo & 0x80)) )
                     return (long)bar;
               }

               ErrorExit ("Error: Unexpected end of file");
               return 0xFFFF;
            }


            //-------------------------------------------------------------------------
            //   GetParam
            //   Called with the input handles
            //   Gets a command paramater

            static char GetParam (ref BinaryReader fileh)
            {
               short? foo;

               if ((foo = fileh.ReadBytes(1)[0]) == null)
                  ErrorExit ("Error: Unexpected end of file\n");
               if ( Convert.ToBoolean(foo & 0x80) )
                  Console.Write("Warning: Paramater with bit 7 set\n");

               return (char)foo;
            }


            //-------------------------------------------------------------------------
            //   CopyEvents
            //   Called with the input and output file handles plus special modifiers
            //   Copies the sound events

            static void CopyEvents (ref BinaryReader inf, ref BinaryWriter outf, ushort tempoconst, int prog, int fixcontrols)
            {
               long? deltat;
               byte eventtime, statusbyte, laststatus = 0, p1, p2 = 0;
               short? bar;
               ulong qt = 0, micro, foo, time = 0, prevtime = 0;

               // loop until the end of the file
               while ( (deltat = ReadVL(ref inf)) != null )  // get the delta time
               {

                  // scale delta t to microseconds
                  micro = ((ulong)deltat * qt) / tempoconst;
                  // get the theoretical time for this event
                  time += micro;

                  // recalculate delta time based on actual time
                  micro = time - prevtime;

                  // convert from microseconds to sound ticks
                  foo = micro / 16000;
                  if ( foo > 255 )
                  {
                     Console.Write("Warning: Delta time too large - truncating\n");
                     foo = 255;
                  }
                  eventtime = (byte)foo;

                  // update prevtime with the actual elapsed time in microseconds
                  micro = foo * 16000;
                  prevtime += micro;

                  // FIXME: JN - if qt is zero, then set the event time to the deltat
                  //    Not sure if this is correct.
                  if (qt == 0)
                  {
                      eventtime = (byte)deltat;
                  }

                  // get the next byte
                  if ( (bar = ReadChar(ref inf)) == -1 )
                     ErrorExit ("Error: Unexpected end of file\n");
                  statusbyte = (byte)bar;
                  p2++;

                  // check for running status byte by checking bit 7
                  if ( !(Convert.ToBoolean(statusbyte & 0x80)) )
                  {
                     statusbyte = laststatus;
                     inf.BaseStream.Position--;    // rewind to read it as a paramater
                  }

                  // copy the command
                  switch ( statusbyte & 0xF0 )
                  {
                     // these 2-paramater commands can be directly copied
                     case 0x80:    // note off
                     case 0x90:    // note on
                     case 0xA0:    // key pressure
                     case 0xE0:    // pitch wheel
                        p1 = (byte)GetParam(ref inf);
                        p2 = (byte)GetParam(ref inf);
                        outf.Write(eventtime);
                        if ( statusbyte != laststatus )
                           outf.Write(statusbyte);
                        outf.Write(p1);
                        outf.Write(p2);
                        break;

                     // these 1-paramater commands can be directly copied
                     case 0xD0:    // channel pressure
                        p1 =(byte) GetParam(ref inf);
                        outf.Write(eventtime);
                        if ( statusbyte != laststatus )
                           outf.Write(statusbyte);
                        outf.Write(p1);
                        break;

                     // the rest are special case commands

                     // check for set MIDI instrument
                     case 0xC0:    // set program
                        p1 = (byte)GetParam(ref inf);
                        if (prog != -1)
                        {
                            p1 = (byte)prog;
                        }

                        outf.Write(eventtime);

                        if (statusbyte != laststatus)
                            outf.Write(statusbyte);

                        outf.Write(p1);
                        break;

                     // check for fix controls
                     case 0xB0:    // set control
                        p1 = (byte)GetParam(ref inf);
                        p2 = (byte)GetParam(ref inf);
                        switch ( p1 )
                        {
                           // the recognized controls
                           case 0x07:
                           case 0x0A:
                           case 0x40:
                           case 0x4B:
                           case 0x4E:
                              outf.Write(eventtime);
                              if ( statusbyte != laststatus )
                                 outf.Write(statusbyte);
                              outf.Write(p1);
                              outf.Write(p2);
                              break;

                           default:
                              if ( !Convert.ToBoolean(fixcontrols) )
                              {
                                 outf.Write(eventtime);
                                 if ( statusbyte != laststatus )
                                    outf.Write(statusbyte);
                                 outf.Write(p1);
                                 outf.Write(p2);
                              }
                            break;
                        }

                        break;

                     case 0xF0:
                        switch ( statusbyte & 0x0F )
                        {
                           // various events which take no data
                           case 0x08:
                           case 0x0A:
                           case 0x0C:
                              break;

                           // system exclusive messages
                           case 0x00:
                           case 0x07:
                              p1 = (byte)GetParam(ref inf);
                              inf.BaseStream.Position += p1; 
                              break;

                           // meta event
                           case 0x0F:
                              p1 = (byte)GetParam(ref inf);
                              switch ( p1 )
                              {
                                 // end of track
                                 case 0x2F:
                                    p1 = 0xFC;
                                    outf.Write(eventtime);
                                    outf.Write(p1);
                                    return;

                                 // microseconds per quarter note
                                 case 0x51:
                                    p1 = (byte)GetParam(ref inf);
                                    if ((bar = ReadChar(ref inf)) == -1)
                                       ErrorExit ("Error: Unexpected end of file\n");
                                    qt = (ushort)bar;
                                    if ((bar = ReadChar(ref inf)) == -1)
                                       ErrorExit ("Error: Unexpected end of file\n");
                                    qt <<= 8;
                                    qt |= (ushort)bar;
                                    if ((bar = ReadChar(ref inf)) == -1)
                                       ErrorExit ("Error: Unexpected end of file\n");
                                    qt <<= 8;
                                    qt |= (ushort)bar;
                                    break;

                                 default:
                                    p1 = (byte)GetParam(ref inf);
                                    inf.BaseStream.Position += p1; 
                                    break;
                              }
                              break;

                           default:
                              ErrorExit ("Error: Unrecognized Fx command\n");
                              break;
                           }
                        break;

                     default:
                        p2++;
                        ErrorExit ("Error: Invalid command\n");
                        break;
                  }

                  laststatus = statusbyte;
               }

               Console.Write("Warning: Unexpected end of file\n");
            }



            //-------------------------------------------------------------------------
            //   ReadMThd
            //   Called with the handle a MIDI file
            //   Reads the header, returning the song tempo

            static ushort ReadMThd (ref BinaryReader filh)
            {
               short foo;
               ushort tempo;
               var header = new byte[] {0x4D, 0x54, 0x68, 0x64, 0x00, 0x00, 0x00, 0x06, 0x00, 0x00, 0x00, 0x01, 0x00, 0x20 };

               byte[] header2 = filh.ReadBytes(12);

               header.ToString();

               if (header.ToString() != header2.ToString())
               {
                   ErrorExit("Error: Invalid file format");
               }


               if ( (foo = ReadChar(ref filh)) == -1 )
                  ErrorExit ("Error: Unexpected end of file");
               tempo = (ushort)(foo << 8);
               if ( (foo = ReadChar(ref filh)) == -1 )
                  ErrorExit ("Error: Unexpected end of file");
               tempo |= (ushort)foo;

               return tempo;
            }



            //-------------------------------------------------------------------------
            //   ReadMTrk
            //   Called with the handle to a new MIDI file
            //   Reads the track header

            static void ReadMTrk (ref BinaryReader outf)
            {
               // I could have just incremented the file pointer, but I may eventually
               //    want to do something with this value
               outf.BaseStream.Position += 8; 
            }



            //-------------------------------------------------------------------------
            //   WriteHeader
            //   Called with the handle a new sound file
            //   Writes the sound header, initializing channels

            static void WriteHeader (ref BinaryWriter fileh, int devices)
            {
               byte foo;

               foo = 0;
               fileh.Write(foo);

               for ( int lcv=0; lcv<16; lcv++)
               {
                   // If no device is specified, default to Adlib (original configuration provided by R.I.)
                   if (devices == 0)
                   {
                       devices = 4;
                   }
              
                  // enable the channel
                  foo = 0x01;
                  fileh.Write(foo);

                  // enable the devices
                  foo = (byte)devices;
                  fileh.Write(foo);
               }
            }


            static void WriteHeaderFromConfigFile(ref BinaryWriter fileh, ref string xmlConfigFile)
            {
                byte[] header = new byte[32];

                XDocument xml = XDocument.Load(xmlConfigFile);

                foreach (XElement element in xml.Descendants("channels"))
                {
                    foreach (XElement childElement in element.Descendants())
                    {
                        int channelNum = Convert.ToInt16(childElement.Attribute("number").Value);
                        int voices = Convert.ToInt16(childElement.Attribute("voices").Value);
                        int devices = Convert.ToInt16(childElement.Attribute("devices").Value);

 
                      //  Console.WriteLine("channeldevices);

                        header[channelNum * 2] = Convert.ToByte(voices);
                        header[channelNum * 2 + 1] = Convert.ToByte(devices);
                    }
                }
                
                // Write out the sound resource type (0 for non-digital, 2 for digital)
                byte foo = 0;
                fileh.Write(foo);

                fileh.Write(header);
            }



            //-------------------------------------------------------------------------
            //   InitControls
            //   Called with the handle a new sound file
            //   Initializes channels

            static void InitControls (ref BinaryWriter fileh)
            {
               byte time = 0, status, p1, p2;

               for ( byte lcv=0; lcv<16; lcv++ )
               {
                  status = (byte)(0xB0 | lcv);
                  p1 = 0x4E;
                  p2 = 8;
                  fileh.Write(time);
                  fileh.Write(status);
                  fileh.Write(p1);
                  fileh.Write(p2);
               }
            }

    }
}
